/* DSP_util.c */
/* $Id: DSP_util.c,v 1.2 1995/09/21 03:51:32 arnaud Exp arnaud $ */

/*********************************************************************
 TCP_talk: real-time sound transmission over a LAN
    
    Copyright (C) 1995  Arnaud Louet

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***********************************************************************/

#include <string.h> /* to access function 'memcpy' */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include "DSP_util.h"
#include <sys/soundcard.h>

int init_DSP( struct DSP_config *param, int read_write)
{
  int filedesc;			/* file descriptor to handle DSP_NAME */
  
  switch (read_write)
    {
    case DSP_READ:
      if ( (filedesc = open(DSP_NAME, O_RDONLY,0)) == -1)
	{
	  perror("init_DSP");
	  return(-1);
	}
      break;
    case DSP_WRITE:
      if ( (filedesc = open(DSP_NAME, O_WRONLY,0)) == -1)
	{
	  perror("init_DSP");
	  return(-1);
	}
      break;
    case DSP_READWRITE:
      if ( (filedesc = open(DSP_NAME, O_RDWR,0)) == -1)
	{
	  perror("init_DSP");
	  return(-1);
	}
      break;
    default:
      fprintf(stderr, "init_DSP: read_write argument is invalid.\n");
      return(-1);
    } 
  
  if(ioctl(filedesc, SOUND_PCM_WRITE_CHANNELS, &(param->channels)) == -1)
    {
      perror("init_DSP");
      return (-1);
    }
  if (ioctl(filedesc, SOUND_PCM_WRITE_BITS, &(param->bits)) == -1)
    {
      perror("init_DSP");
      return (-1);
    }
  if (ioctl(filedesc, SOUND_PCM_WRITE_RATE, &(param->rate)) == -1)
    {
      perror("init_PCM");
      return (-1);
    }  
  if (ioctl(filedesc, SOUND_PCM_SUBDIVIDE, &(param->subdivide)) == -1)
    {
      perror("init_PCM");
      return (-1);
    }
  return filedesc;
}

      

void probe_DSP( struct DSP_config *requested, struct DSP_config *actual)
{
  int filedesc;
    
  if (actual == NULL || requested == NULL )
    exerror("probe_DSP");

  memcpy(actual, requested, sizeof (struct DSP_config));

  if ((filedesc = init_DSP(actual, DSP_READ)) == -1)
    {
      fprintf(stderr,"check_DSP failed.\n");
      exit(EXIT_FAILURE);
    }

  close(filedesc);
  
}

int DSP_ready()
{
  int filedesc;
  
  if ( ( filedesc = open(DSP_NAME, O_RDONLY, 0)) == -1 )
    return 0;			/* ie DSP not ready */
    
  if ( close(filedesc) == -1 )
    exerror("DSP_ready");
  
  return 1;			/* DSP ready */
}

int check_DSP( struct DSP_config *requested)
{
  struct DSP_config obtained;
  
  probe_DSP( requested, &obtained );
  if (obtained.rate < (0.9 * requested->rate) )
    return 0;
  if (obtained.rate > (1.1 * requested->rate) )
    return 0;
  if (obtained.channels != requested->channels)
    return 0;
  if (obtained.bits != requested->bits )
    return 0;

  return 1;
}

void discard_DSP(int DSP_filedesc)
{
  /* discards the DSP drivers's DMA buffer  */
  /* I'm not sure of the value returned by ioctl */
  if (ioctl(DSP_filedesc, SOUND_PCM_RESET, 0) == -1)
    {
      perror("discard_DSP");
      exit(EXIT_FAILURE);
    }
}

void flush_DSP(int DSP_filedesc)
{
  /* discards the DSP drivers's DMA buffer  */
  /* I'm not sure of the value returned by ioctl */
  if (ioctl(DSP_filedesc, SOUND_PCM_SYNC, 0) == -1)
    {
      perror("flush_DSP");
      exit(EXIT_FAILURE);
    }
}




  
    




























