/* DSP_util.h  header for the library file DSP_util*/
/* $Id: DSP_util.h,v 1.1 1995/05/24 16:21:24 arnaud Exp $ */ 

#define DSP_NAME "/dev/dsp"

struct DSP_config		/* contains the parameters passed to */
{				/* the function init_DSP   */
  int rate;			/* sampling rate [Hz] */
  int channels;			/* mono or stereo */
  int bits;			/* bits per sanples e.g: 8, 12, 16 */
  int subdivide;		/* size of buffer between DSP and computer */
};


#define DSP_mono 1		/* coressponds to the field channels of  */
#define DSP_stereo 2		/* struct DSP_config */

#define FULL_BUFFER 1		/* coresponds to the field subdivide of  */
#define HALF_BUFFER 2		/* struct DSP_config */
#define QUARTER_BUFFER 4	/* FULL_BUFFER= 0.5 to 1 second of sound */

#define DSP_READ 0		/* corresponds to second argument of */
#define DSP_WRITE 1		/* init_DSP */
#define DSP_READWRITE 2

/* ---------------------------------------------------------------------- */
/* init_DSP initialises the sound card's DSP  for read or write operation */
/*          returns a file-descriptor on /dev/dsp in case of success      */
/*          returns -1 in case of a failure                               */
/* struct DSP_config is mofified if the parameters are not supported by   */
/*                     the sound board                                    */
/* ---------------------------------------------------------------------- */
int init_DSP( struct DSP_config *param, int read_write);

/* Tests if DSP supports the sound parameters PARAM              */
/* returns the actual supported values with the pointer 'actual' */
/* DSP_config structure                                          */
       
void probe_DSP( struct DSP_config *requested, struct DSP_config *actual);



int DSP_ready();
				/* This function returns a true value (1) */
				/* if the loacal DSP is ready to be used  */
				/* (i.e.) is not used by another application */
				/* it returns 0 otherwise */


				
/* check_DSP tries to set the DSP to the `requested` values */
/* returns 1 if it is possible, 0 if the requested values */
/* could not be obtained. */
/* N.B: this function uses probe_DSP */

int check_DSP( struct DSP_config *requested);

/* discards the DMA buffer of the DSP driver                   */
/* This allows to switch immediately from read to write        */
/* Information can get lost ususally between .25s and 2seconds */
void discard_DSP(int DSP_filedesc);

/* flushes the DMA buffer of the DSP driver            */
/* ie: continue playback/record until DMA buffer is empty */
/* This allows to switch from read to write */
/* NO INFORMATION IS LOST */
void flush_DSP(int DSP_filedesc);

















