/* $Id: TCP_talk_clnt.c,v 1.7 1995/08/24 19:20:23 arnaud Exp arnaud $ */


/*********************************************************************
 TCP_talk: real-time sound transmission over TCP/IP
    
    Copyright (C) 1995  Arnaud Louet

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***********************************************************************/



#include "Copyright.h"
#include "debug.h"
#include "talk.h"
#include "sock_utils.h"
#include "DSP_util.h"
#include "misc.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>


 
int Sound_Connect(char *host, char *service, struct DSP_config *Requested_Sound_Param, int *DSPdesc, int *sockdesc, unsigned long flag_compression);

				/* tries to connect to another machine    */
				/* in order to start a conversation       */
				/* return value: STT_READY if successful  */
                                /*               STT_DSP failed because   */
                                /*                one of the DSPs is busy */
		                /*               STT_NOT_READY otherwise  */
				/* returns a file desc. on /dev/dsp and   */
				/*  a socket descripor as last arguments  */


void usage(void)
{
  fprintf( stderr, "Usage: TCP_talk_clnt [-p PORT] [-s SAMPLING RATE] [-b BITS] [-B small|medium|large] [-C blanks|824|sinus] [-S] [-P] host\n");
  exit(EXIT_FAILURE);
}


int Sound_Connect(char *host, char *service, struct DSP_config *Requested_Sound_Param, int *DSPdesc, int *sockdesc, unsigned long flag_compression)
{
  int n;
  int status_code;
  char control[2];

  struct DSP_config *Server_Sound_Param; /* i.e. remote machine */
  struct DSP_config *Local_Sound_Param;
    
  Local_Sound_Param= (struct DSP_config *) malloc( sizeof (struct DSP_config));
  Server_Sound_Param= (struct DSP_config *) malloc(sizeof (struct DSP_config));

  

  if ( !DSP_ready() )
    {
      fprintf( stderr, "Error opening DSP\nProgram aborted.\n");
      return STT_DSP_NOT_READY;
    }

  *sockdesc= connectTCP(host,service); 
	
  blk_read(*sockdesc, &status_code, sizeof status_code);
  
  if ( status_code != STT_READY)
    {
      return status_code;
    }
  probe_DSP(Requested_Sound_Param, Local_Sound_Param);

#ifdef DEBUG
  fprintf( stderr, "Parameters asked for: rate: %d Hz, channels: %d, bits/sample: %d.\n",
	 Requested_Sound_Param->rate,
	 Requested_Sound_Param->channels,
	 Requested_Sound_Param->bits);
#endif
  
#ifdef DEBUG
  fprintf( stderr, "Parameters supported by local board: rate: %d Hz, channels: %d, bits/sample: %d.\n",
	 Local_Sound_Param->rate,
	 Local_Sound_Param->channels,
	 Local_Sound_Param->bits);
#endif

  blk_write(*sockdesc, Local_Sound_Param, sizeof (struct DSP_config));
    
  blk_read(*sockdesc, Server_Sound_Param, sizeof( struct DSP_config));
  
  blk_write(*sockdesc, &flag_compression, sizeof flag_compression);
  
  fprintf( stderr, "Parameters retained by server: rate: %d Hz, channels: %d, bits/sample: %d.\n",
	 Server_Sound_Param->rate,
	 Server_Sound_Param->channels,
	 Server_Sound_Param->bits);
  
  if ( (*DSPdesc = init_DSP(Server_Sound_Param, DSP_READWRITE)) == -1 )
    {
      fprintf( stderr, "init_DSP failed\nProgram aborted.\n");
      exit( EXIT_FAILURE);
    }

  free(Server_Sound_Param);
  free(Local_Sound_Param);
  
  return(STT_READY);
}

      
int main(int argc, char* argv[])

{
  struct DSP_config Sound_Parameters;
  int socket_desc;		/* socket descriptor */
  int DSP_filedesc;		/* file descriptor for the DSP */
  int status_code;
  int state;
   int status;
  extern char buffer[BLOCK_SIZE];
  unsigned long flag_compression;
  
  int index;
  char c;
  
  char *host;
  char *port;
  
  float ratio_send;
  float ratio_receive;

  int send_len;  

  port = (char *) malloc( 30 * sizeof (char));
  port = DEFAULT_PORT;         /* default well-known port */

  Sound_Parameters.bits= 8 ;		/* default values: 8bits mono 8kHz */
  Sound_Parameters.channels = 1;
  Sound_Parameters.rate = 8000;
  Sound_Parameters.subdivide = FULL_BUFFER; /* sets DMA buffer to 1/4s  */
                                               /* approx of recording time */

  flag_compression = FLAG_NO_COMPRESSION;
 
  
   while ((c = getopt (argc, argv, "p:s:b:B:C:I:SPT:")) != -1)
    switch (c)
      {
      case 'p':
	port = optarg;
	break;
      case 's':
	Sound_Parameters.rate = atoi(optarg);
	break;
      case 'B':
	 if ( !strcmp(optarg,"small") )
	   Sound_Parameters.subdivide = QUARTER_BUFFER;
	 else if ( !strcmp(optarg, "medium"))
	   Sound_Parameters.subdivide = HALF_BUFFER;
	 else if ( !strcmp(optarg, "large"))
	   Sound_Parameters.subdivide = FULL_BUFFER;
	 else
	   usage();
	 break;
       case 'C':			/* compression type */
	 if ( !strcmp(optarg, "blanks"))
	   flag_compression |= FLAG_COMPRESS_BLANKS;
	 else if ( !strcmp(optarg, "824"))
	   flag_compression |= FLAG_COMPRESS_824;
	 else if ( !strcmp(optarg, "sinus"))
	   {
	     flag_compression |= FLAG_COMPRESS_SINUS;
	     init_sin_tbl();
	   }
	 else 
	   {
	      usage();
	   }
	 break;
       case 'S':
	 flag_compression |= FLAG_SCRAMBLE;
	 break;
       case 'P':
	 flag_compression |= FLAG_PERMUTE;
	 break;
       case 'T':			/* trigger level for blank compression */
	 Trigger = atoi(optarg);
	 break;
       case 'I':			/* interval in samples for blank compr. */
	 Interval = atoi(optarg);
	 break;
       case 'b':
	 Sound_Parameters.bits = atoi(optarg);	
	 break;
       default:
	 usage();
      }
   
  if( optind != (argc-1))
    {
       usage;
    }
  host = argv[optind];	     /* name or IP address of the host to connect to */
  

  if ( (status_code = Sound_Connect(host, port, &Sound_Parameters, &DSP_filedesc, &socket_desc, flag_compression)) != STT_READY )
    {
      switch (status_code)
	{
	case STT_DSP_NOT_READY: 
	  fprintf( stderr, "Sound card of remote machine is not ready.\nProgram aborted.\n");
	  exit(EXIT_FAILURE);
	  break;
	case STT_ALREADY_CONNECTED:
	  fprintf( stderr, "Other host is already connected to someone else.\nProgram aborted\n");
	  exit(EXIT_FAILURE);
	default:
	  fprintf( stderr, "Sound connect failed.\nProgram aborted.\n");
	  exit(EXIT_FAILURE);
	}
    }

  disable_input_buffering();	/* disables echo+ buffering of the input  */
  stdin_flush();		/* terminal */

  
  /* setsockopt(socket_desc, SOL_SOCKET, SO_SNDLOWAT, &send_len, sizeof (send_len)); */
  
  /* main loop= actual sound conversation */
  for ( ; ; )
    {
      stdin_flush();
      fprintf(stderr, "Now sending \n");
      while ( (state = send_block_TCP( socket_desc, DSP_filedesc, flag_compression )) == SR_CONTINUE )
	;
#ifdef DEBUG
      fprintf(stderr, "receive block returned %d after WHILE loop\n", state);
#endif
      if ( state == SR_DISCONNECT )
	break;
      flush_DSP(DSP_filedesc);
      fprintf(stderr, "Now receiving \n");
      while ( (state = receive_block_TCP( socket_desc, DSP_filedesc, flag_compression )) == SR_CONTINUE )
	;
      
      if ( state == SR_DISCONNECT )
	break;
      flush_DSP(DSP_filedesc);
    }
    
  ratio_send = (float)global_count_send / global_compressed_send ;
  ratio_receive = (float)global_count_receive / global_compressed_receive ;
#ifdef DEBUG
  fprintf(stderr,"Efficiency of compression (send): %f .\n", ratio_send);
  fprintf( stderr, "Mean tranfer rate (send): %f kbits/s.\n", 64/ratio_send);
  fprintf(stderr,"Efficiency of compression (receive): %f .\n", ratio_receive);
  fprintf( stderr, "Mean tranfer rate (receive): %f .\n", 64/ratio_receive);
#endif
}


