/* $id: talk.c,v 1.9 1995/06/07 13:00:50 arnaud Exp arnaud $ */

/*******************************************************************
 TCP_talk: real-time sound transmission over TCP/IP
    
    Copyright (C) 1995  Arnaud Louet

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***********************************************************************/


#include "Copyright.h"
#include "debug.h"
#include "talk.h"
#include "misc.h"
#include "lookup.h"

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>


#define DELTA_MAX 7
#define DELTA_MIN -8

char *passwd1="Raymondldsfjfjdslk;gdfhwgjkdfhljfdshkfh;,f;jkhgjl;khfj'kj;klhgk;lgfhjklghjfh;kl;ghmbhjgklfhdbgmjfldhjgl;hjgfhdfjshgdfjmngfmnsdgbwuergwebuigbtruitrgbgyuigbtrugtrbyu";
char *passwd2="Raymondldsfjfjdslk;gdfhwgjkdfhljfdshkfh;,f;jkhgjl;khfj'kj;klhgk;lgfhjklghjfh;kl;ghmbhjgklfhdbgmjfldhjgl;hjgfhdfjshgdfjmngfmnsdgbwuergwebuigbtruitrgbgyuigbtrugtrbyu";

static unsigned int sin_tbl[3218];

unsigned long int global_count_send = 0;
unsigned long int global_count_receive = 1;
unsigned long int global_compressed_send = 0;
unsigned long int global_compressed_receive = 1;

char buffer[BLOCK_SIZE];
char compress_buffer[5][2*BLOCK_SIZE]; /* up to 5 compression/encryption */
                                       /* can be chained */
char decompress_buffer[5][2*BLOCK_SIZE];

const int SR_CONNECT_REQUEST = 3;
const int SR_DISCONNECT = 2;
const int SR_FLIP = 1;
const int SR_CONTINUE = 0;

struct curve
{
  unsigned short position;
  unsigned char extremum;
};


int send_block_TCP( int sock_fd, int DSP_fd, unsigned long flag_compression )
{
   unsigned char *buf_out;
   int out_size;
   
   char c;
   int i;
   
   blk_read(DSP_fd, buffer, BLOCK_SIZE);
   
   if (buffer[0] == 0)
     buffer[0] = 2;
   
   for ( i=0 ; i<BLOCK_SIZE; i++)
     {                               /* suppresses values that coud be  */
	if( buffer[i] == MARKER )    /* interpreted as control codes */
	buffer[i]=2;
     }
   
   compress( flag_compression, buffer, BLOCK_SIZE, &buf_out, &out_size);

   global_count_send += BLOCK_SIZE;
   global_compressed_send += out_size;
   
   sock_write( sock_fd, buf_out, out_size );
   
   c = get_key();
   if ( c == ' ')
     {
       buffer[0] = 0;
       buffer[1] = SR_FLIP;
       sock_write( sock_fd, buffer, 2 );
       i = SR_FLIP;
       return i;
     }
   else if ( c == 'q'|| c == 'Q')			
     {
       buffer[0] = 0;
       buffer[1] = SR_DISCONNECT;
       sock_write( sock_fd, buffer, 2 );
       close( sock_fd );
       close( DSP_fd );
#ifdef DEBUG
       fprintf(stderr,"Quit in send_block_TCP\n");
#endif
       i= SR_DISCONNECT;
#ifdef DEBUG
       fprintf(stderr, "value returned: %d\n", i);
#endif
       return i;
     }
   else 
     {
       i = SR_CONTINUE;
       return i;
     }
 }


int receive_block_TCP( int sock_fd, int DSP_fd, unsigned long flag_compression)
{
  int in_size;
  unsigned char *out_buf;
  int out_size;
  
  int i;
  char c;
  
  sock_read(sock_fd, buffer, &in_size);
  global_compressed_receive += in_size;
  if ( buffer[0] == 0 )	   /* a control datagram has been received */
    {
      if ( buffer[1] == SR_DISCONNECT)
	 {
	    close(sock_fd);
	    close(DSP_fd);
	    return SR_DISCONNECT;
	 }
	else if (buffer[1] == SR_FLIP)
	 {
	    return SR_FLIP;
	 }
	else
	 {
	    printf("Unknown control code received.\nProgram aborted.\n");
	    exit(EXIT_FAILURE);
	 }
    }

   
   decompress(flag_compression, buffer, in_size, &out_buf, &out_size);

   global_count_receive += out_size;
   

   
   blk_write( DSP_fd, out_buf, out_size);
   return SR_CONTINUE;
}

void compress(unsigned long flags, unsigned char *buf_in, int in_len, unsigned char **buf_out, int *out_len)
{
  static int d=2;
  static int e=0;
  int i=-1;
  int tempo_len = in_len;
  char *tempo_buf = buf_in;
      
  if( flags == FLAG_NO_COMPRESSION)
    {
      *buf_out = buf_in;
      *out_len = in_len;
      return;
    }
  
  if ( flags & FLAG_COMPRESS_BLANKS )
    {
      *buf_out = compress_buffer[++i];
      compress_blank( tempo_buf, tempo_len, *buf_out, out_len);
      tempo_len = *out_len;
      tempo_buf = *buf_out;
    }
  if ( flags & FLAG_COMPRESS_824 )
    {
      *buf_out = compress_buffer[++i];
      compress_824( tempo_buf, tempo_len, *buf_out, out_len);
      tempo_len = *out_len;
      tempo_buf = *buf_out;
    }
  if ( flags & FLAG_COMPRESS_SINUS )
    {
      *buf_out = compress_buffer[++i];
      compress_sinus( tempo_buf, tempo_len, *buf_out, out_len);
      tempo_len = *out_len;
      tempo_buf = *buf_out;
    }
  if ( flags & FLAG_SCRAMBLE )
    {			      
      *buf_out = compress_buffer[++i];
      scramble( tempo_buf, tempo_len, *buf_out, out_len, passwd1);
      tempo_len = *out_len;
      tempo_buf = *buf_out;
    }
  if ( flags & FLAG_PERMUTE ) 
    {			      
      *buf_out = compress_buffer[++i];
      *buf_out[0] = tempo_buf[0];
      permute( tempo_buf+1, tempo_len-1, (*buf_out)+1, 4 /*d*/);
      *out_len = tempo_len;
      tempo_len = *out_len;
      tempo_buf = *buf_out;
      d++;	
      d = (d%3)+1;
    }
  if ( flags & FLAG_PERMUTE_2 ) 
    {			      
      *buf_out = compress_buffer[++i];
      *buf_out[0] = tempo_buf[0];
      permute_2( tempo_buf+1, tempo_len-1, (*buf_out)+1,out_len , lookup_table[e], 8);
      tempo_len = *out_len;
      tempo_buf = *buf_out;
      e++;
      if ( e == 6)
	e= 0;
    }
  
}

void decompress(unsigned long flags, unsigned char *buf_in, int in_len, unsigned char **buf_out, int *out_len)
{
  static int d=2;
  static int e=0;
  int i = -1;
  int tempo_len = in_len;
  char *tempo_buf = buf_in;
  
  if( flags == FLAG_NO_COMPRESSION)
    {
      *buf_out = buf_in;
      *out_len = in_len;
      return;
    }

  if ( flags & FLAG_PERMUTE_2 ) 
    {			      
      *buf_out = decompress_buffer[++i];
      *buf_out[0] = tempo_buf[0];
      unpermute_2( tempo_buf+1, tempo_len-1, (*buf_out)+1, out_len, lookup_table[e], 8);
      tempo_len = *out_len;
      tempo_buf = *buf_out;
      e++;
      if ( e == 6)
	e= 0;
    }
  if ( flags & FLAG_PERMUTE ) 
    {			      
      *buf_out = decompress_buffer[++i];
      *buf_out[0] = tempo_buf[0];
      unpermute( tempo_buf+1, tempo_len-1, (*buf_out)+1, 4/*d*/);
      *out_len = tempo_len;
      tempo_len = *out_len;
      tempo_buf = *buf_out;
      d++;	
      d = (d%3)+1;
    }

  if ( flags & FLAG_SCRAMBLE )
    {				
      *buf_out = decompress_buffer[++i];
      unscramble( tempo_buf, tempo_len, *buf_out, out_len, passwd2);
      tempo_len = *out_len;
      tempo_buf = *buf_out;
    }
  if ( flags & FLAG_COMPRESS_824 )
    {
      *buf_out = decompress_buffer[++i];
      decompress_824( tempo_buf, tempo_len, *buf_out, out_len);
      tempo_len = *out_len;
      tempo_buf = *buf_out;
    }
  if ( flags & FLAG_COMPRESS_BLANKS )
    {
      *buf_out = decompress_buffer[++i];
      decompress_blank( tempo_buf, tempo_len, *buf_out, out_len);
      tempo_len = *out_len;
      tempo_buf = *buf_out;
    }
  if ( flags & FLAG_COMPRESS_SINUS )
    {
      *buf_out = decompress_buffer[++i];
      decompress_sinus( tempo_buf, tempo_len, *buf_out, out_len);
      tempo_len = *out_len;
      tempo_buf = *buf_out;
    }
}



void compress_824(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len)
{

  int i;
  int j = 1;
  int k;
  int c;
  register int delta;
  register int delta_2;
  register int carry = 0;
  char first;
  
  out_buf[0] = in_buf[0];	/* the 1st char is directly copied to output */

  for( i=1 ;  i < (in_len-1) ; /* empty */ )
    {
      k = 0;
      while ( in_buf[i] == MARKER ) /* Preserves control codes from blank compression */
	{
	  k++;
	  memcpy(out_buf+j, in_buf+i, 1+sizeof(unsigned short));
	  j+= (1 + sizeof(unsigned short));
	  i+= (1 + sizeof(unsigned short));
	  if( i >= (in_len-1))
	    break;		/* break from the while loop */
	}
      if( i >= (in_len-1))
	break;			/* break from the for loop */
      if ( k != 0)
	carry += in_buf[i] - in_buf[i - 1 - k * (1 + sizeof(unsigned short))];
      else
	carry += in_buf[i] - in_buf[i-1];
      
      if ( carry > DELTA_MAX)
	delta = DELTA_MAX;
      else if( carry >= DELTA_MIN )
	delta = carry;
      else
	delta = DELTA_MIN;
      carry -= delta;
      
      i++;

      k = 0;
      while ( in_buf[i] == MARKER )
	{
	  k++;
	  memcpy(out_buf+j, in_buf+i, 1+sizeof(unsigned short));
	  j+= (1 + sizeof(unsigned short));
	  i+= (1 + sizeof(unsigned short));
	  if( i >= in_len)
	    break;		/* break from the `while` loop */
	}
      if( i >= in_len)
	break;			/* break from the `for` loop */
      if ( k != 0)
	carry += in_buf[i] - in_buf[i - 1 - k * (1 + sizeof(unsigned short))];
      else
	carry += in_buf[i] - in_buf[i-1];


      if ( carry > DELTA_MAX)
	delta_2 = DELTA_MAX;
      else if( carry >= DELTA_MIN )
	delta_2 = carry;
      else
	delta_2 = DELTA_MIN;
      carry -= delta_2;
      
      delta += 8;		/* convert the deltas so that the */
      delta_2 += 8;		/* negative values fit in a char  */
      
      out_buf[j] = (char) ( (delta << 4) + delta_2 );
      if ( out_buf[j] == MARKER )
	{
	  out_buf[j]=0x10;	/* this replaces delta=-8 delta_2=-7 (0x01) */
	}                       /* by delta=-7 delta_2=-8 (0x10) */
      j++;			/* in order to prevent generating the MARKER */
      i++;			/* control code used by blank compression  */
    }
  *out_len = j;
}

void decompress_824(unsigned char *in_buf, int in_len,unsigned char *out_buf, int *out_len)
{
  register int previous;
  register char delta;
  int i;
  int j = 1;
  
  out_buf[0] = previous = in_buf[0];
  for( i=1 ; i < in_len ; i++ )
    {
      while ( in_buf[i] == MARKER )
	{
	  memcpy(out_buf+j, in_buf+i, 1+sizeof(unsigned short));
	  j+= (1 + sizeof(unsigned short));
	  i+= (1 + sizeof(unsigned short));
	  if( i >= in_len)
	    break;		/* break from the `while` loop */
	}
      if( i >= in_len)
	break;			/* break from the `for` loop */
      
      delta = (in_buf[i] >> 4) & 0x0f;	/* 4 MSB of in_buf[i] */
      delta -= 8;		/* get back to a signed value */
      previous += delta;
      out_buf[j] = previous;
      if ( out_buf[j] == MARKER )
	out_buf[j] = 2;
      j++;
      delta = in_buf[i] & 0x0f; /* 4 less significant bits of in_buf[i] */
      delta -= 8;
      previous += delta;
      out_buf[j] = previous;
      if ( out_buf[j] == MARKER )
	out_buf[j] = 2;
      j++;
    }
  *out_len = j;
}

int Trigger = 1;
int Interval = 4;		

void compress_blank(unsigned char *in_buf, int in_len,unsigned char *out_buf, int *out_len)
{
  const unsigned char Trigger_min = ( 128 - Trigger );
  const unsigned char Trigger_max = ( 128 + Trigger );

  register bool flag_blank = FALSE;
  unsigned short count = 0;
  int i;
  int j = 1;
  
  out_buf[0] = in_buf[0];
  for ( i = 1 ; i < in_len ; i++ )
    {
      if ( flag_blank == FALSE )
	{			/* compression not triggered */
	  out_buf[j] = in_buf[i/*j*/];
	  if (  ( in_buf[i] < Trigger_min) || ( in_buf[i] > Trigger_max) )
	    count = 0;
	  count ++;
	  j++;
	  
	  if(count == Interval)
	    {
	      j -= (Interval - 2);
	      flag_blank = TRUE;
	    }
	}
      else			/* compression triggered */
	{
	  if (  (in_buf[i] < Trigger_min) || (in_buf[i] > Trigger_max) )
	    {
	      out_buf[j] = MARKER;
	      j++;
	      count --;
	      memcpy( out_buf+j , &count, sizeof count);
	      j += sizeof count;
	      flag_blank = FALSE;
	      count = 0;
	    }
	  count ++;
	}
    } /* end of loop */
  
  if (flag_blank == TRUE )
    {
      out_buf[j] = MARKER;		/* may be wrong byffer */
      j++;
      --count;
      memcpy( out_buf+j, &count, sizeof count);
      j+=sizeof count;
      *out_len = j;
    }
  else
    {
      *out_len = j;
    }
}



void decompress_blank(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len)
{
  unsigned char Trigger_min = ( 128 - Trigger );
  unsigned char Trigger_max = ( 128 + Trigger );

  int i;
  int j=0;
  unsigned short count;
  
  for( i=0 ; i < in_len ; i++)
    {
      if( in_buf[i] == MARKER )
	{
	  i++;
	  memcpy( &count, in_buf + i, sizeof count);
	  memset(out_buf+j, 128, count);
	  j+=count;
	  i+=(sizeof count) - 1; /* not shure -1 is necessary */
	}
      else
	{
	  out_buf[j] = in_buf[i];
	  j++;
	}
    }
  *out_len = j;
}

void init_sin_tbl(void)
{
  int i;

  fprintf(stderr,"Initialising  sinus table [                                         ]");
  fprintf(stderr,"\rInitialising  sinus table [");
  for( i=0; i<3218; i++)	/* 3127 = 2*pi << 9 */
    {
      sin_tbl[i]=(unsigned int)(sin((float)i/512)*512);
      if( i%80 == 0)
	fprintf(stderr, ".");
    }
  fprintf(stderr, "]\n");
}


unsigned char 
f( unsigned int start, unsigned int end, unsigned int extremum, unsigned int x)
{
  unsigned char result;
                          /* formula using only integer operations */
                          /*   pi*512 = 1608  */
  
  if (extremum >= 128) 
    result = ( (sin_tbl[(1608u*(x-start))/(end-start)] * \
		(extremum-128u)) >>9) +128u;
  else
    result = 128u - ( (sin_tbl[(1608u*(x-start))/(end-start)] * \
		       (128u-extremum)) >>9); 
  return (unsigned char)result;   
} 



 
/* unsigned char
f( unsigned int start, unsigned int end, unsigned int extremum, unsigned int x)
{
  unsigned char result;
  
  result =  (unsigned char) (128 + (sin(PI*((float)x-start)/(end-start)) * ((float)extremum-128)));
  
  return result;   
} 
*/




void compress_sinus(unsigned char *in_buf, int in_len, unsigned char* out_buf, int *out_len) 
{
  bool Flag_positive;
  int i;
  int j = 1;
  unsigned char xtrem = in_buf[1];
  struct curve temp;
  
  out_buf[0] = in_buf[0];
  if ( (unsigned char)in_buf[1] >= 128 )
    {
      Flag_positive = TRUE;
    }
  else
    {
      Flag_positive = FALSE;
    }
  
  for ( i=2 ; i<in_len ; i++)
    {
      if (Flag_positive == TRUE)
	{
	  if ( (in_buf[i] < 128) && (temp.position != (i-1)) )
	    {               /* switch from positive to negative signal */
	      Flag_positive = FALSE;
	      temp.position = i;
	      temp.extremum = xtrem;
	      memcpy( out_buf+j, &temp, sizeof temp);
	      j += sizeof temp;
	      xtrem = in_buf[i];
	    }
	  else
	    {
	      xtrem = max(xtrem, in_buf[i]);
	    }
	}
      else			/* switch from negative to positive signal */
	{
	  if( in_buf[i] >= 128  && (temp.position != (i-1)) )
	    {                     /* swtich from - to + */
	      Flag_positive = TRUE;
	      temp.position = i;
	      temp.extremum = xtrem;
	      memcpy(out_buf+j, &temp, sizeof temp);
	      j +=  sizeof temp;
	      xtrem = in_buf[i];
	    }
	   else
	    {
	      xtrem = min( xtrem, in_buf[i]);
	    }
	}
    }
  temp.position = i;		/* nb this may introduce distortion */
  temp.extremum = xtrem; 	/* the effect is to terminate the   */
				/* buffer by a sinusoid ending on a */
				/* zero level. */
  memcpy(out_buf+j, &temp, sizeof temp);
  j +=  sizeof temp;
  
  *out_len = j;		
}                       

void decompress_sinus(unsigned char *in_buf, int in_len, unsigned char* out_buf, int *out_len)
{
  struct curve current;
  struct curve previous= { 1, 128};
  int i;
  int j;
  
  out_buf[0] = in_buf[0];
  for( i=1; i < in_len; i+=sizeof(struct curve))
    {
      memcpy( &current, in_buf+i, sizeof current);
      for( j=previous.position; j<current.position; j++)
	{
	  out_buf[j] = f(previous.position, current.position, current.extremum, j);
	}
      previous.position = current.position;
      previous.extremum = current.extremum;
    }
  *out_len = j;
}

void scramble(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len, char *passwd)
{
  char pass[500];
  int passwd_len = strlen(passwd);
  int i;
  int j = 0;
  unsigned char tempo;
    
  out_buf[0] = in_buf[0];
  strcpy(pass, passwd);		/* make a local copy of passwd        */
                                /* this coppy can be modified without */     
                                /* altering the original              */
  for ( i=0; i< passwd_len; i++) 
    {				
      pass[i]=(unsigned char)(((unsigned int)pass[i] * 11) % 256);
    }      
  
  for( i=1; i<in_len; i++)
    {
      tempo = in_buf[i];
      in_buf[i] = (in_buf[i] >> 4) & 0x0f;
                                /* the 4 MSB are pushed to the left */
      in_buf[i] |= tempo << 4;	/* in the end the 4MSB are inverted  */
				/* with the 4 LSB */
      out_buf[i] = in_buf[i] ^ pass[j++]; /* X-OR the passwd w/ data */
      if ( j>=passwd_len)
	j = 0;
    }
  *out_len = in_len;
}

void unscramble(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len, char *passwd)
{
  char pass[500];
  int passwd_len = strlen(passwd);
  int i;
  int j = 0;
  unsigned char tempo;
  unsigned int t;
  
  out_buf[0] = in_buf[0];
  strcpy(pass, passwd);  

  for ( i=0; i< passwd_len; i++)
    {
      pass[i]=(unsigned char)(((unsigned int)pass[i] * 11) % 256);
    }      
    
  for( i=1; i<in_len; i++)
    {
      out_buf[i] = in_buf[i] ^ pass[j++];
      
      tempo = out_buf[i];        /* invert 4 MSB and 4LSB  */
      out_buf[i] =(out_buf[i] >> 4) & 0x0f;	
      out_buf[i] |= tempo << 4;	
			
      if ( j>=passwd_len)
	j = 0;
    }

  *out_len = in_len;
}

void permute(unsigned char *in_buf, int in_len, unsigned char *out_buf, int nIteration)
{
  if (nIteration == 1)
    {
      memcpy( out_buf, in_buf, in_len);
      permutation( out_buf, in_len, in_len/2);
    }
  else 
    {
      permute( in_buf, in_len/2, out_buf+in_len/2,nIteration -1);
      permute( in_buf+in_len/2, in_len-in_len/2, out_buf, nIteration -1);
    }       
}

void unpermute(unsigned char *in_buf, int in_len, unsigned char *out_buf, int nIteration)
{
  if (nIteration == 1)
    {
      memcpy( out_buf, in_buf, in_len);
      permutation( out_buf, in_len, in_len - in_len/2 );
    }
  else 
    {
      unpermute( in_buf, in_len/2, out_buf+in_len/2, nIteration -1);
      unpermute( in_buf+in_len/2, in_len-in_len/2,out_buf, nIteration -1);
    }       
}

void permutation( unsigned char *buffer, int buffer_len, int cut)
{
  char temp_buf[buffer_len];
  if (cut <= 0 || cut >= buffer_len)
    {
      fprintf(stderr, "Last argument of permute is out of range\nProgram aborted\n");
      exit(EXIT_FAILURE);
    }
  memcpy( temp_buf, buffer+cut, buffer_len - cut);
  memmove( buffer + buffer_len-cut, buffer, cut);
  memcpy( buffer, temp_buf, buffer_len - cut);
}


void permute_2(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len, unsigned short *lookup_tbl, unsigned short ltlen)
{
  unsigned int sub_block_size = in_len/ltlen;
  unsigned short i;
  

  for( i=0 ; i < ltlen ; i++)
    memcpy(out_buf + lookup_tbl[i]*sub_block_size, \
	   in_buf  + i * sub_block_size, \
	   sub_block_size);
  if( ltlen * sub_block_size < in_len)
    memcpy(out_buf + ltlen*sub_block_size,
	   in_buf + ltlen*sub_block_size,
	   in_len - ltlen * sub_block_size);
  *out_len = in_len;
}




void unpermute_2(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len, unsigned short *lookup_tbl, unsigned short ltlen)
{
  static bool first_call = TRUE;
  unsigned short reverse_lt[ltlen];
  unsigned int sub_block_size = in_len/ltlen;
  unsigned short i;
  
  
  for ( i=0 ; i< ltlen; i++)	/* generates the reverse lookup table */
    reverse_lt[lookup_tbl[i]]=i;
  
  for( i=0 ; i < ltlen ; i++)
    memcpy(out_buf + reverse_lt[i]*sub_block_size, \
	   in_buf  + i * sub_block_size, \
	   sub_block_size);
  if( ltlen * sub_block_size < in_len)
    memcpy(out_buf + ltlen*sub_block_size,
	   in_buf + ltlen*sub_block_size,
	   in_len - ltlen * sub_block_size);
  *out_len = in_len;
      
}




