/* $Id: talk.h,v 1.10 1995/09/16 21:55:39 arnaud Exp arnaud $ */


/*********************************************************************
 TCP_talk: real-time sound transmission over TCP/IP
    
    Copyright (C) 1995  Arnaud Louet

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***********************************************************************/



/* defines common to UDP_talk_server and UDP_talk_srvr */
/* DSP can be assimilated to sound-card in the rest of this file */

#include <math.h>
#include <string.h>		/* for (de)compress_blank */

#define DEFAULT_PORT "25000"
#define TCP_BUFF_SIZE 64000	/* TCP layer buffer size: 8000= 1s of sound  */
				/* if 8bits/sample and 8kHz are chosen       */

#define BLOCK_SIZE 1025         /* a power of 2+1 is safer in order to       */
				/* achive correct permutation algorithm      */
                                /* if permute has a depth 10, the buflen     */
				/* shoud be at least 2^10+1=1025             */
extern unsigned long global_count_send;
extern unsigned long global_count_receive;
extern unsigned long global_compressed_send;
extern unsigned long global_compressed_receive;

extern char *passwd1, *passwd2;

extern char buffer[BLOCK_SIZE];

                                /* the following  macros describe the*/
#define STT_ALREADY_CONNECTED 2 /* state of the connection           */
#define STT_DSP_NOT_READY 1	/* DSP of one of the hosts not ready */
#define STT_NOT_READY -1        /*remote host not ready (undefined reason)*/
#define STT_READY 0		/* connection ready (successful)     */

extern const int SR_CONNECT_REQUEST; /* only used in UDP version  */
extern const int SR_DISCONNECT;   /* values returned by the send or receive */
extern const int SR_FLIP;	  /* procedures                             */
extern const int SR_CONTINUE;	  /* actually not a PDU but put here for */
				  /*  convenience.                       */

#define FLAG_NO_COMPRESSION 0
#define FLAG_COMPRESS_824 1
#define FLAG_COMPRESS_BLANKS 2
#define FLAG_COMPRESS_SINUS 4
#define FLAG_SCRAMBLE 8
#define FLAG_PERMUTE 16
#define FLAG_PERMUTE_2 32

extern int Trigger;		/* trigger levek for blank compression */
extern int Interval; 		/* interval in samples to trigger blank */
				/* compression */
#define MARKER 1		/* marker for blank compression */


int send_block_TCP( int sock_fd, int DSP_fd, unsigned long flag_compression );
int receive_block_TCP(int sock_fd, int DSP_fd, unsigned long flag_compression );

                               /* these two functions call the aproporiate   */
			       /* compression functions and return a pointer */
			       /* on a buffer where the compressed data block*/
			       /* the lenth of the output data is in         */
			       /*  *out_len                                  */
void compress(unsigned long flags, unsigned char *buf_in, int in_len, unsigned char **buf_out, int *out_len);
void decompress(unsigned long flags, unsigned char *buf_in, int in_len, unsigned char **buf_out, int *out_len);


void compress_824(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len);
void decompress_824(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len);

void compress_blank(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len);
void decompress_blank(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len);

unsigned char f( unsigned int start, unsigned int end, unsigned int extremum, unsigned int x);

void compress_sinus(unsigned char *in_buf, int in_len, unsigned char* out_buf, int *out_len);
void decompress_sinus(unsigned char *in_buf, int in_len, unsigned char* out_buf, int *out_len);
void init_sin_tbl(void);


void scramble(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len, char *passwd);
void unscramble(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len, char *passwd);

void permute(unsigned char *in_buf, int in_len, unsigned char *out_buf, int nIteration);
void unpermute( unsigned char *in_buf, int in_len, unsigned char *out_buf, int nIteration);

void permutation( unsigned char *buffer, int buffer_len, int cut);
				/* divides a buffer in two parts and     */
				/* permutes the two parts                */
				/* 1st part: buf[0]..buf[cut-1]          */
				/* 2nd part: buf[cut]..buf[buffer_len-1] */


void
permute_2(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len, unsigned short *lookup_tbl, unsigned short ltlen);

void
unpermute_2(unsigned char *in_buf, int in_len, unsigned char *out_buf, int *out_len, unsigned short *lookup_tbl, unsigned short ltlen);
















