/* sock_utils.c */
/* part of this code is copyrighted by Comer, Stevens and Prentice Hall,
 * here is their copyright notice 
 * 
 *  Portions of this software are covered by the following copyright:
 * Copyright (c) 1993, 1994 Douglas E. Comer, David L. Stevens, and
 * Prentice Hall, Inc. All rights reserved.
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the copyright holders. The names of the copyright holders
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty. The authors
 * assume no liability for damages incidental or consequential, nor is the
 * software warranted for correctness or suitability for any purpose.
 * Portions of this software are documented in the book: 
 * Comer, D. E. and D. L. Stevens [1993], 'Internetworking with TCP/IP Vol 3: 
 * Client-Server Programming and Applications', Prentice-Hall,
 * Englewood Cliffs, New Jersey.
 * This software may not be sold or published in printed form without written
 * permission from the copyright holders. */


#include "sock_utils.h"
#include <stdlib.h>
#include <strings.h>
#include <sys/types.h>		/* contains definition of size_t */
#include <sys/socket.h>

#include <arpa/inet.h>   /* defines: type in_addr, protocols: TCP,UDP...  */
                          /* macro: INADDR_ANY, INADDR_BROADCAST,          */
			  /*        INADDR_NONE, INADDR_LOOPBACK...        */
                          /* functions: htonl, ntohl, htons, ntohs         */

#include <netdb.h>        /* define: type struc hostent, servent, protoent */
			  /* functions: gethostbyname, getprotobyname,     */
			  /*            getservbyname, gethostbyaddr,      */
			  /*            etc...*/


/* ------------------------------------------------------------------------- 
 *   connectTCP - connect to a specified TCP service on a specified host 
 * ------------------------------------------------------------------------- 
 */
int connectTCP(char* host, char* service)
{
  return connectsock( host, service, "tcp");
}

/* -------------------------------------------------------------------------
 *   connectUDP - connect to a specified UDP service on a specified host
 * -------------------------------------------------------------------------
 */
int connectUDP(char* host, char* service)
{
  return connectsock( host, service, "udp");
  
}
/*--------------------------------------------------------------------------*/
/* Connectsock-  allocates & connect a socket using TCP or UDP              */
/* -------------------------------------------------------------------------*/

int connectsock(char* host,char* service,char* protocol )
{
  struct hostent *phe;		/* pointer to name resolution table entry */
  struct servent *pse;		/* pointer to service information entry   */
  struct protoent *ppe;		/* pointer to protocol information entry  */
  struct sockaddr_in *sin;	/* an internet endpoint address           */
  int s;			/* socket descriptor(similar to file desc)*/
  int type;			/* socket type */
  
  if ((sin = (struct sockaddr_in *) calloc(1, sizeof (struct sockaddr_in))) == 0) 
    {           
      perror("Connectsock: cannot allocate memory for sin..."); 
      exit(1);
    }
  sin->sin_family = AF_INET;


  /* Map service name to port nuber */
  if (pse = getservbyname(service, protocol) )
    sin->sin_port = pse->s_port;
  else		
    if ( (sin->sin_port = htons((u_short)atoi(service))) == 0 )
      {
	perror("connectsock: cannot get port number of service...");
	exit(1);
      }
  
  
  /* map host name to IP address */
  if (phe = gethostbyname(host) )
    bcopy(phe->h_addr, (char *) &sin->sin_addr, phe->h_length);
  else if
    ( (sin->sin_addr.s_addr = inet_addr(host)) == INADDR_NONE )
      {
	perror("connectsock: cannot get host entry...");
	exit(1);
      }
  
  /* map protocol name to protocol number */
  if ((ppe = getprotobyname(protocol)) == 0)
    {
      perror("connectsock: can't get protocol entry...");
      exit(1);
    }

  /* Use protocol to choose a socket type */
  if ( strcmp(protocol,"udp") == 0)
    type = SOCK_DGRAM;
  else
    type = SOCK_STREAM;
  
  /* allocate a socket */
  s = socket(PF_INET, type, ppe->p_proto);
  if (s < 0)
    {
      perror("connectsock: can't create socket...");
      exit(1);
    }
  
  /* connect the socket */
  if ( connect(s, (struct sockaddr *) sin, sizeof (struct sockaddr_in)) < 0)
    {
      perror("connectsock: can't establish connection...");
      exit(1);
    }
  
  return s;
}


/* ---------------------------------------------------------------------- */
/* passiveUDP- creates a passive socket for use in a UDP server           */
/*      service: service associated with the desired port                 */
/* ---------------------------------------------------------------------- */
int passiveUDP( char* service )
{
  return passivesock(service, "udp", 0);
}

/* ---------------------------------------------------------------------- */
/* passiveTCP-creates a passive socket for unse in a TCP server           */
/*       qlen: length of connection reauest queue                         */
/*       service: service associated with the desired port                */
/* ---------------------------------------------------------------------- */
int passiveTCP( char* service, int qlen )
{
  return passivesock(service, "tcp", qlen);
}



/* ------------------P O R T B A S E -------------------------------------*/
/* portbase is non zero if a new version of the service has to be tested  */
/* while the old working version is running on the host. ex portbase=1024 */

u_short portbase = 0;

/* ---------------------------------------------------------------------- */
/*  passivesock - allocate & bind a server socket using TCP or UDP        */
/* ---------------------------------------------------------------------- */


int passivesock( char* service, char* protocol, int qlen)
{
  struct servent *pse;		/* pointer to service info entry */
  struct protoent *ppe;		/* pointer to protocol information entry */
  struct sockaddr_in *sin;	/* an internet endpoint address */
  int s;			/* socket descriptor */
  int type;			/* socket type */
  
 if ((sin = (struct sockaddr_in *) calloc(1, sizeof(struct sockaddr_in))) == 0)
    {
      perror("passivesock: cannot allcate memory for variable sin...");
      exit(1);
    }
  sin->sin_family = AF_INET;
  sin->sin_addr.s_addr = INADDR_ANY; /*listens to all IP addresses of host */
  
  /* map service name to port number */
  if ( pse = getservbyname( service, protocol) )
    sin->sin_port = htons(ntohs((u_short)pse->s_port) + portbase);
  else if ( (sin->sin_port = htons((u_short)atoi(service))) == 0 )
    {
      perror("passivesock: can't determine port number...");
      exit(1);
    }

  /* map protocol name to protocol number */
  if ( (ppe = getprotobyname(protocol)) == 0 )
    {
      perror("passivesock: can't determine protocol number...");
      exit(1);
    }
  
  /* use protocol to choose a socket type */
  if ( strcmp(protocol, "udp") == 0)
    type = SOCK_DGRAM;
  else
    type = SOCK_STREAM;
  
  /* allocate a socket */
  s = socket(PF_INET, type, ppe->p_proto);
  if ( s < 0)
    {
      perror("passivesock: can't create socket...");
      exit(1);
    }
  
  /* bind the socket to its well-known port */
  if (bind(s, (struct sockaddr *)sin, sizeof(*sin)) < 0)
    {
      perror("passivesock: can't bind socket to well known port...");
      exit(1);
    }

  /* for tcp (SOCK_STREAM) sockets only: use listen to queue requests */
  if (type == SOCK_STREAM && listen(s, qlen) < 0)
    {
      perror("passivesock: can't listen to well known port...");
      exit(1);
    }
  
  return s;
}

  
      
      
/* sets the buffers length of the TCP layer buffer   */
/* len_TX: pointer on length for the transmit buffer */
/* len_RX: pointer on length fo the receive buffer   */
void set_TCP_bufflen( int sock, int *len_TX, int *len_RX)
{
  if( *len_TX != 0 )
    {
      if( setsockopt(sock, SOL_SOCKET, SO_SNDBUF, len_TX, sizeof (int)) == -1 )
	exerror("setsockopt in adjust_TCP_buflen");
    }
  if ( *len_RX != 0)
    {
      if( setsockopt(sock, SOL_SOCKET, SO_RCVBUF, len_RX, sizeof (int)) == -1 )
	exerror("setsockopt in adjust_TCP_buflen");
    }
}

void sock_read( int sock_fd, void *buf, unsigned int *count)
{
  blk_read( sock_fd, count, sizeof (unsigned int));
  blk_read( sock_fd, buf, *count);
}


void sock_write( int sock_fd, void *buf, unsigned int count)
{
  blk_write( sock_fd, &count, sizeof count);
  blk_write( sock_fd, buf, count);
}

