.TH WAVPLAY/WAVREC 1 LOCAL
.SH VERSION
Version 0.21w1

.SH NAME
wavplay - Plays .wav files

wavrec - Records .wav files
.SH SYNOPSIS
.B wavplay 
[-h] [-q] [-s rate] [-S] [-t secs] [-b bits] [-k key] [-r] [infiles...]

.B wavplay 
[-k key] [-r] [-l] [-u] [-L] [-U]

.B wavrec
[-h] [-q] [-s rate] [-S] [-t secs] [-b bits] [-k key] [-r] [outfile]

.SH DESCRIPTION 

\fIwavplay\fP plays sound files in the Windows 3.1 .WAV format through
your sound card. Without any options given on the command line, all
required settings like sampling rate, stereo or mono mode, bits per
sample (8, 12, 16) will be supplied from the header of the given
file(s). You can override these settings by using the command line
options.

\fIwavrec\fP samples data coming in from your external devices like a
microphone or CD player and writes it to standard output or the
designated filename. With no command line options, the default settings
will be used.

\fIRecording Defaults:\fP

.RS
.B -s rate
: 22050 Hz

.B -t secs
: 10 seconds

.B -b bits
: 8 Bits

.B -S
: mono (no -S)
.RE

When no file names are given for \fIwavplay\fP, the input is assumed to
come from standard input (when \fIwavplay\fP needs input). When no file
names are given for \fIwavrec\fP, the output is sent to standard output.

Messages are sent to standard error (see \fI-q\fP option).

Commands \fIwavplay\fP and \fIwavrec\fP employ locking semaphores to
prevent multiple processes running the same commands from playing or
recording at the same time respectively. This is particularly useful if
you allow \fIwavplay\fP to run in the background during system startup
for example. This mode of operation is shown in the first form of the
\fIwavplay\fP example in this file.

The second \fIwavplay\fP command form shows only locking options.
Options \fI-l, -u, -L\fP and \fI-U\fP cause \fIwavplay\fP to perform
only the corresponding \fIlock\fP or \fIunlock\fP on the corresponding
\fIplay\fP or \fIrecord\fP locking semaphores. Their purpose is to allow
the shell script writer the access to the same locking mechanism used 
by \fIwavplay\fP and \fIwavrec\fP commands when other sound tools are
used.

The \fI-k\fP option allows the administrator further flexibility,
since this permits the maintenance of multiple locking sets for multiple
sound cards, or even other resources.

.SH OPTIONS
.B -S
: sets stereo-mode

.B -s rate
: sets the sample frequency in Hz

.B -t secs
: sets the time limit in seconds

.B -b bits
: sets the sample size to 8, 12 or 16 bits. The Soundblaster card
supports only 8 bits.

.B -q
: quiet mode for no message output.

.B -h
: gives a help text.

.B -k key
: specifies a decimal numeric value to be used as the unique IPC key 
for the creation and locating of the locking semaphores. Normally, the
compiled in default is used and this option is not required.

.B -r
: forces the removal of the existing locking semaphore set, and the
recreation of a new semaphore set. This can be used if you suspect that the
existing set has been messed up some how.

.B -l
: locks the \fIplay\fP semaphore. This can be used to gate access to the
sound driver in shell scripts, using this command's builtin locking
semaphores. \fIwavplay\fP when used with this option only performs the
lock operation, and no playing is done in this mode (this assumes that
the shell script will launch another sound program).

.B -u
: unlocks the \fIplay\fP semaphore. Again \fIwavplay\fP does not play
any sounds in this mode, but simply undoes the effect of the \fI-l\fP
option.

.B -L
: like the \fI-l\fP, this option causes the \fIrecord\fP semaphore to
be locked.

.B -U
: this undoes the effect of the \fI-L\fP option.

.SH FILES
.B $PATH/wavplay 
: .wav player or locking semaphore manager

.B $PATH/wavrec
: record .wav files.

.B /dev/dsp
: this is determined by the compile manifest AUDIODEV

.SH IPC Key

.B 0x33333333
: this is the compiled in hexidecimal default for the commands. However
this is determined by compile time manifest AUDIOLCK which may be
modified and different for your site.

.SH DIAGNOSTICS
If an error occurs, a message is written to standard error and an
exit(-1) is taken. This results in the shell returning $? value of 255.

If \fIwavplay\fP is used for locking or unlocking only, the return value
is 0 if successful, or 1 if the semaphore operation failed for some
reason (-r reset could not be accomplished due to permissions is a
frequent cause).

Warnings may be issued if the options conflict with settings in
the .wav file.

.SH BUGS

This should probably be under \fIfeatures\fP: when \fI-r\fP is used to
remove and recreate the locking semaphores, the operation can be denied
by the kernel and the \fIwavplay\fP program will exit. The most likely
cause of this is that the semaphores are owned by another user, and you
lack the privileges to remove it.

The  semaphore locks can also get out of sync if the commands are killed
with \fI-SIGKILL\fP (kill -9). Also note that the semaphores are created
with world  access  0666, so on a hostile system (with students?), it is
possible for other hostile programs to monkey with these semaphores and
leave the locks in a state that will hang otherwise legitimate runs.

.SH EXAMPLES
.B $ cat sample.wav | wavplay 

plays sample.wav

.B $ wavplay sample.wav	

plays this sample too.

.B $ wavrec -S -t 10 -s 10000 test.wav

records data in stereo with a sampling rate of 10000 Hz for 10 seconds

.B $ wavplay -l ; other_cmd ; wavplay -u

locks the \fIplay\fP semaphore, runs \fIother_cmd\fP, and finally
unlocks the \fIplay\fP semaphore.

.B $ wavplay -L ; other_cmd ; wavplay -U

does the same thing, but uses the \fIrecord\fP locking semaphore
instead.

.B $ wavplay -r file.wav

plays file.wav but forces a removal and recreation of the locking
semaphore set.

.B $ wavplay -k 34567 file.wav

plays file.wav, but uses 34567 as the IPC key for the locking
semaphores.

.SH SEE ALSO
This work was derived from wavplay021.tar.z and patch file
wavplay021_patch.tar.gz. See these for the original programs and the
xplay(1) program.

.SH AUTHORS
.B Some code used from Audio Utilities for Linux 2.0 (c) Liam Corner

.B Andre Fuechsel (C) FF-Soft '93

.B e-mail: 
af1@irz.inf.tu-dresden.de

The modifications in this release (wavplay021w1), and the writer of \fIlocks.c\fP:

.B Warren W. Gay VE3WWG

.B email:
bx249@freenet.toronto.on.ca, wwg@ica.net
